#include "aim.h"

/* At one point, we are running a single global work group (or "total
 * work" or "NDRange") on the GPU, which consists of N independent
 * tasks ("(local) work group"), each operating on its own image.
 * Within each work group, we run M threads ("work item") collaborating
 * on the image; not all threads might run at once, they are divided
 * to warps which take turns on the multiprocessors while others get
 * stalled waiting for the memory.
 *
 * The hardware is essentially "transparently SIMD" through the warp;
 * if one thread in the warp does something different from the others,
 * all other threads in the warp stall. */

/* Within each task, the threads may use (size-limited to few kb)
 * local memory. Just declare a variable (in any scope) with the
 * __local keyword. Also, you can make a workgroup barrier
 * (ensure all threads within a workgroup reach a certain point before
 * any goes on) by calling barrier(CLK_LOCAL_MEM_FENCE);. That's useful
 * to make sure the threads all initialized their bit of local memory
 * before continuing in the execution. */

/* Refer to the number of current work group using get_group_id(0).
 * Refer to the number of work item (thread) within the work group
 * using get_local_id(0). Refer to the global number of thread (use
 * if you ignore work groups) using get_group_id().
 * You can also use get_global_size(0), get_local_size(0),
 * get_num_groups(0). Also, you can use 2^n-dimensional vectors as work
 * coordinates instead of single values.
 * A useful idiom is to perform once-only initializations (e.g. setup
 * of shared variables) using:
 * if (!get_local_id(0)) { ... }; barrier(CLK_LOCAL_MEM_FENCE); */

 /* TODO: Wave. */
 /* TODO: Rescale 1:8. */
 /* TODO: Rotate. */
__kernel void
exercise(__global struct image *img_in, __global unsigned char *bitmap_in,
	 __global struct image *img_out, __global unsigned char *bitmap_out,
	 unsigned int bitmapsize)
{
	/* Set up our proper state. */
	/* TODO: Make this __local and do it only once? */
	__global struct image *img = &img_in[get_group_id(0)];
	img_out[get_global_id(0)] = *img;
	//int rows = img->rows;
	//int rowbytes = image_rowbytes(img);
	__global unsigned char *ibitmap = &bitmap_in[get_group_id(0) * bitmapsize];
	__global unsigned char *obitmap = &bitmap_out[get_group_id(0) * bitmapsize];

	/* Now, divide the picture to many segments; within one
	 * segment, each thread in our work group will take care of
	 * a single pixel. */
	int segsize = get_local_size(0);
	/* XXX: assert(bitmapsize % segsize == 0) */

	for (int segstart = 0; segstart < bitmapsize; segstart += segsize) {
		/* TODO: Work on ints instead. */
		/* TODO: Vectorize. */
		int i = segstart + get_local_id(0);
		obitmap[i] = ~ibitmap[i];
	}
}
